@@ -17,9 +17,12 @@ Usage:
 
 """
 
+import sys
+import glob
 import os.path
 import ctypes
 import ctypes.util
+import threading
 
 from ctypes import c_char_p, c_int, c_size_t, c_void_p
 
@@ -31,30 +34,35 @@ class Magic:
 
     """
 
-    def __init__(self, mime=False, magic_file=None, mime_encoding=False):
+    def __init__(self, mime=False, magic_file=None, mime_encoding=False,
+                 keep_going=False):
         """
         Create a new libmagic wrapper.
 
         mime - if True, mimetypes are returned instead of textual descriptions
         mime_encoding - if True, codec is returned
         magic_file - use a mime database other than the system default
-
+        keep_going - don't stop at the first match, keep going
         """
         flags = MAGIC_NONE
         if mime:
             flags |= MAGIC_MIME
         elif mime_encoding:
             flags |= MAGIC_MIME_ENCODING
+        if keep_going:
+            flags |= MAGIC_CONTINUE
 
         self.cookie = magic_open(flags)
 
         magic_load(self.cookie, magic_file)
 
+        self.thread = threading.currentThread()
 
     def from_buffer(self, buf):
         """
         Identify the contents of `buf`
         """
+        self._thread_check()
         return magic_buffer(self.cookie, buf)
 
     def from_file(self, filename):
@@ -62,44 +70,64 @@ class Magic:
         Identify the contents of file `filename`
         raises IOError if the file does not exist
         """
-
+        self._thread_check()
         if not os.path.exists(filename):
             raise IOError("File does not exist: " + filename)
 
         return magic_file(self.cookie, filename)
 
+    def _thread_check(self):
+        if self.thread != threading.currentThread():
+            raise Exception('attempting to use libmagic on multiple threads will '
+                            'end in SEGV.  Prefer to use the module functions '
+                            'from_file or from_buffer, or carefully manage direct '
+                            'use of the Magic class')
+
     def __del__(self):
-        # during shutdown magic_close may have been cleared already
+        # no _thread_check here because there can be no other
+        # references to this object at this point.
+
+        # during shutdown magic_close may have been cleared already so
+        # make sure it exists before using it.
+
+        # the self.cookie check should be unnessary and was an
+        # incorrect fix for a threading problem, however I'm leaving
+        # it in because it's harmless and I'm slightly afraid to
+        # remove it.
         if self.cookie and magic_close:
             magic_close(self.cookie)
             self.cookie = None
 
-_magic_mime = None
-_magic = None
 
-def _get_magic_mime():
-    global _magic_mime
-    if not _magic_mime:
-        _magic_mime = Magic(mime=True)
-    return _magic_mime
-
-def _get_magic():
-    global _magic
-    if not _magic:
-        _magic = Magic()
-    return _magic
+instances = threading.local()
 
 def _get_magic_type(mime):
-    if mime:
-        return _get_magic_mime()
-    else:
-        return _get_magic()
+    i = instances.__dict__.get(mime)
+    if i is None:
+        i = instances.__dict__[mime] = Magic(mime=mime)
+    return i
 
 def from_file(filename, mime=False):
+    """"
+    Accepts a filename and returns the detected filetype.  Return
+    value is the mimetype if mime=True, otherwise a human readable
+    name.
+    
+    >>> magic.from_file("testdata/test.pdf", mime=True)
+    'application/pdf'
+    """
     m = _get_magic_type(mime)
     return m.from_file(filename)
 
 def from_buffer(buffer, mime=False):
+    """
+    Accepts a binary string and returns the detected filetype.  Return
+    value is the mimetype if mime=True, otherwise a human readable
+    name.
+
+    >>> magic.from_buffer(open("testdata/test.pdf").read(1024))
+    'PDF document, version 1.2'
+    """
     m = _get_magic_type(mime)
     return m.from_buffer(buffer)
 
@@ -117,12 +145,14 @@ if dll:
 if not libmagic or not libmagic._name:
     import sys
     platform_to_lib = {'darwin': ['/opt/local/lib/libmagic.dylib',
-                                  '/usr/local/lib/libmagic.dylib',
-                                  '/usr/local/Cellar/libmagic/5.10/lib/libmagic.dylib'],
+                                  '/usr/local/lib/libmagic.dylib'] +
+                       # Assumes there will only be one version installed
+                       glob.glob('/usr/local/Cellar/libmagic/*/lib/libmagic.dylib'),
                        'win32':  ['magic1.dll']}
     for dll in platform_to_lib.get(sys.platform, []):
         try:
             libmagic = ctypes.CDLL(dll)
+            break
         except OSError:
             pass
 
@@ -132,13 +162,26 @@ if not libmagic or not libmagic._name:
 
 magic_t = ctypes.c_void_p
 
-def errorcheck(result, func, args):
-    err = magic_error(args[0])
-    if err is not None:
+def errorcheck_null(result, func, args):
+    if result is None:
+        err = magic_error(args[0])
         raise MagicException(err)
     else:
         return result
 
+def errorcheck_negative_one(result, func, args):
+    if result is -1:
+        err = magic_error(args[0])
+        raise MagicException(err)
+    else:
+        return result
+    
+
+def coerce_filename(filename):
+    if filename is None:
+        return None
+    return filename.encode(sys.getfilesystemencoding())
+
 magic_open = libmagic.magic_open
 magic_open.restype = magic_t
 magic_open.argtypes = [c_int]
@@ -155,26 +198,30 @@ magic_errno = libmagic.magic_errno
 magic_errno.restype = c_int
 magic_errno.argtypes = [magic_t]
 
-magic_file = libmagic.magic_file
-magic_file.restype = c_char_p
-magic_file.argtypes = [magic_t, c_char_p]
-magic_file.errcheck = errorcheck
+_magic_file = libmagic.magic_file
+_magic_file.restype = c_char_p
+_magic_file.argtypes = [magic_t, c_char_p]
+_magic_file.errcheck = errorcheck_null
 
+def magic_file(cookie, filename):
+    return _magic_file(cookie, coerce_filename(filename))
 
 _magic_buffer = libmagic.magic_buffer
 _magic_buffer.restype = c_char_p
 _magic_buffer.argtypes = [magic_t, c_void_p, c_size_t]
-_magic_buffer.errcheck = errorcheck
-
+_magic_buffer.errcheck = errorcheck_null
 
 def magic_buffer(cookie, buf):
     return _magic_buffer(cookie, buf, len(buf))
 
 
-magic_load = libmagic.magic_load
-magic_load.restype = c_int
-magic_load.argtypes = [magic_t, c_char_p]
-magic_load.errcheck = errorcheck
+_magic_load = libmagic.magic_load
+_magic_load.restype = c_int
+_magic_load.argtypes = [magic_t, c_char_p]
+_magic_load.errcheck = errorcheck_negative_one
+
+def magic_load(cookie, filename):
+    return _magic_load(cookie, coerce_filename(filename))
 
 magic_setflags = libmagic.magic_setflags
 magic_setflags.restype = c_int
